package com.fjhx.mes.service.production.impl;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.util.ObjectUtil;
import com.baomidou.dynamic.datasource.annotation.DS;
import com.baomidou.dynamic.datasource.annotation.DSTransactional;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.fjhx.common.enums.PushBusinessTypeEnum;
import com.fjhx.common.utils.Assert;
import com.fjhx.file.utils.ObsFileUtil;
import com.fjhx.item.entity.product.po.ProductInfo;
import com.fjhx.item.service.product.ProductInfoService;
import com.fjhx.mes.entity.production.dto.ProductionTaskDetailDto;
import com.fjhx.mes.entity.production.dto.ProductionTaskDetailSelectDto;
import com.fjhx.mes.entity.production.po.*;
import com.fjhx.mes.entity.production.vo.ProductionTaskDetailRecordVo;
import com.fjhx.mes.entity.production.vo.ProductionTaskDetailVo;
import com.fjhx.mes.entity.work.po.WorkOrder;
import com.fjhx.mes.entity.work.po.WorkOrderProductionProcesses;
import com.fjhx.mes.entity.work.vo.WorkOrderVo;
import com.fjhx.mes.mapper.production.ProductionTaskDetailMapper;
import com.fjhx.mes.service.production.*;
import com.fjhx.mes.service.work.WorkOrderProductionProcessesService;
import com.fjhx.mes.service.work.WorkOrderService;
import com.fjhx.socket.core.PushTypeEnum;
import com.fjhx.socket.core.WebSocketPush;
import com.fjhx.wms.entity.stock.emums.StockWaitType;
import com.fjhx.wms.entity.stock.po.StockWait;
import com.fjhx.wms.entity.stock.po.StockWaitDetails;
import com.fjhx.wms.service.stock.StockWaitDetailsService;
import com.fjhx.wms.service.stock.StockWaitService;
import com.ruoyi.common.exception.ServiceException;
import com.ruoyi.common.utils.SecurityUtils;
import com.ruoyi.common.utils.wrapper.IWrapper;
import com.ruoyi.common.utils.wrapper.SqlField;
import com.ruoyi.system.utils.UserUtil;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.math.BigDecimal;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;


/**
 * <p>
 * 生产任务明细 服务实现类
 * </p>
 *
 * @author
 * @since 2023-05-18
 */
@Service
public class ProductionTaskDetailServiceImpl extends ServiceImpl<ProductionTaskDetailMapper, ProductionTaskDetail> implements ProductionTaskDetailService {
    @Autowired
    private ProductionTaskDetailRecordService productionTaskDetailRecordService;
    @Autowired
    private ProductionTaskService productionTaskService;
    @Autowired
    private ProductionPlanService productionPlanService;
    @Autowired
    private WorkOrderService workOrderService;
    @Autowired
    private WorkOrderProductionProcessesService workOrderProductionProcessesService;
    @Autowired
    private StockWaitService stockWaitService;
    @Autowired
    private StockWaitDetailsService stockWaitDetailsService;
    @Autowired
    private ProductInfoService productInfoService;
    @Autowired
    private ProductionProcessesService productionProcessesService;
    @Autowired
    private ProductionTaskProcessesDetailService productionTaskProcessesDetailService;

    @Override
    public Page<ProductionTaskDetailVo> getPage(ProductionTaskDetailSelectDto dto) {
        IWrapper<ProductionTaskDetail> wrapper = getWrapper();
        return pageCom(wrapper, dto);
    }

    @Override
    public Page<ProductionTaskDetailVo> circulationPage(ProductionTaskDetailSelectDto dto) {
        IWrapper<ProductionTaskDetail> wrapper = getWrapper();
        //过滤负责人为自己的任务
        wrapper.eq(ProductionTaskDetail::getPersonLiableId, SecurityUtils.getUserId());
        //过滤未在流转的记录
        wrapper.eq(ProductionTaskDetail::getCirculationStatus, 0);
        //过滤未完成的记录
        wrapper.eq(ProductionTaskDetail::getFinishStatus, 0);
        return pageCom(wrapper, dto);
    }

    /**
     * 分页公共方法
     */
    Page<ProductionTaskDetailVo> pageCom(IWrapper<ProductionTaskDetail> wrapper, ProductionTaskDetailSelectDto dto) {
        //关键字搜索
        if (ObjectUtil.isNotEmpty(dto.getKeyword())) {
            List<Long> productIds = productInfoService.listObject(ProductInfo::getId, q -> q.like(ProductInfo::getName, dto.getKeyword()));
            wrapper.and(q -> q.in(ProductionTaskDetail::getProductId, productIds).or().like(ProductionTaskDetail::getProductSn, dto.getKeyword()));
        }

        wrapper.orderByDesc("ptd", ProductionTaskDetail::getId);
        wrapper.groupBy("ptd.id");
        Page<ProductionTaskDetailVo> page = this.baseMapper.getPage(dto.getPage(), wrapper);
        List<ProductionTaskDetailVo> records = page.getRecords();
        if (ObjectUtil.isEmpty(records)) {
            return new Page<>();
        }
        //赋值流转人名称
        UserUtil.assignmentNickName(records, ProductionTaskDetailVo::getCirculationUserId, ProductionTaskDetailVo::setCirculationUserName);

        //赋值产品名
        productInfoService.attributeAssign(records, ProductionTaskDetailVo::getProductId, (item, productInfo) -> {
            item.setProductName(productInfo.getName());
        });
        //赋值当前工序名 目标工序名
        List<Long> productionTaskIds = records.stream().map(ProductionTaskDetail::getProductionTaskId).collect(Collectors.toList());
        IWrapper<ProductionTaskDetail> wrapper1 = IWrapper.getWrapper();
        wrapper1.in("pt", ProductionTaskDetail::getId, productionTaskIds);
//        wrapper1.groupBy("wo.id");
        List<WorkOrderVo> workOrderByProductionTask = baseMapper.getWorkOrderByProductionTask(wrapper1);
        Map<Long, WorkOrderVo> WorkOrderMap = workOrderByProductionTask.stream().collect(Collectors.toMap(WorkOrderVo::getProductionTaskId, Function.identity()));
        List<Long> workIds = workOrderByProductionTask.stream().map(WorkOrderVo::getId).collect(Collectors.toList());

        //获取工序id列表
        List<Long> productionProcessesIds = records.stream().map(ProductionTaskDetail::getProductionProcessesId).collect(Collectors.toList());
        productionProcessesIds.addAll(records.stream().map(ProductionTaskDetail::getNextProductionProcessesId).collect(Collectors.toList()));
        productionProcessesIds.addAll(records.stream().map(ProductionTaskDetail::getPreviousProcessesId).collect(Collectors.toList()));
        //获取基本工序map
        Map<Long, String> stringMap = productionProcessesService.mapKV(ProductionProcesses::getId, ProductionProcesses::getName,
                q -> q.in(ProductionProcesses::getId, productionProcessesIds));
        //获取定制工序map
        Map<Long, String> workOrderProductionPMap = workOrderProductionProcessesService.mapKV(WorkOrderProductionProcesses::getId,
                WorkOrderProductionProcesses::getName, q -> q.in(WorkOrderProductionProcesses::getWorkOrderId, workIds));

        for (ProductionTaskDetailVo record : records) {
            WorkOrderVo workOrderVo = WorkOrderMap.get(record.getProductionTaskId());
            if ("1".equals(workOrderVo.getIsCustomized())) {
                //如果是定制
                String name = workOrderProductionPMap.get(record.getProductionProcessesId());
                String nextName = workOrderProductionPMap.get(record.getNextProductionProcessesId());
                String previousName = workOrderProductionPMap.get(record.getPreviousProcessesId());
                record.setProductionProcessesName(name);
                record.setNextProductionProcessesName(nextName);
                record.setPreviousProcessesName(previousName);
            } else {
                //不是定制
                String name = stringMap.get(record.getProductionProcessesId());
                String nextName = stringMap.get(record.getNextProductionProcessesId());
                String previousName = stringMap.get(record.getPreviousProcessesId());
                record.setProductionProcessesName(name);
                record.setNextProductionProcessesName(nextName);
                record.setPreviousProcessesName(previousName);
            }
        }

        return page;
    }

    /**
     * 任务流转
     */
    @Override
    @DSTransactional
    public void circulation(ProductionTaskDetailDto dto) {
        ProductionTaskDetail productionTaskDetail = this.getById(dto.getId());

        //修改流转状态
        productionTaskDetail.setCirculationStatus(1);
        //赋值流转人id
        productionTaskDetail.setCirculationUserId(SecurityUtils.getUserId());

        //赋值上一道工序 为当前工序
        productionTaskDetail.setPreviousProcessesId(productionTaskDetail.getProductionProcessesId());

        //将下一工序赋值到当前工序
        productionTaskDetail.setProductionProcessesId(productionTaskDetail.getNextProductionProcessesId());
        //清空下一工序
        productionTaskDetail.setNextProductionProcessesId(-1l);
        //查找下一工序相关信息 如果有则赋值
        ProductionTask productionTask = productionTaskService.getById(productionTaskDetail.getProductionTaskId());
        ProductionPlan productionPlan = productionPlanService.getById(productionTask.getProductionPlanId());
        WorkOrder workOrder = workOrderService.getById(productionPlan.getWorkOrderId());
        List<WorkOrderProductionProcesses> productionProcessesList = workOrderService.getProductionProcessesList(workOrder);
        //赋值下一工序id
        for (int i = 0; i < productionProcessesList.size(); i++) {
            WorkOrderProductionProcesses wopp = productionProcessesList.get(i);
            if (wopp.getId().equals(productionTaskDetail.getProductionProcessesId())) {
                //默认赋值空 如果有下一节点赋值下一节点id
                if ((i + 1) < productionProcessesList.size()) {
                    WorkOrderProductionProcesses nextProductionProcesses = productionProcessesList.get(i + 1);
                    productionTaskDetail.setNextProductionProcessesId(nextProductionProcesses.getId());
                    break;
                }
            }
        }

        //设置当前工序状态未开始
        productionTaskDetail.setProcessesStatus(0);
        //清空负责人
        productionTaskDetail.setPersonLiableId(-1L);

        this.updateById(productionTaskDetail);

        //创建任务明细记录
        ProductionTaskDetailRecord productionTaskDetailRecord = new ProductionTaskDetailRecord();
        productionTaskDetailRecord.setProductionTaskDetailId(productionTaskDetail.getId());
        productionTaskDetailRecord.setProductionProcessesId(productionTaskDetail.getProductionProcessesId());
        productionTaskDetailRecord.setPersonLiableId(SecurityUtils.getUserId());
        productionTaskDetailRecord.setReceivedTime(productionTaskDetail.getReceivedTime());
        productionTaskDetailRecordService.save(productionTaskDetailRecord);

        //保存附件
        ObsFileUtil.saveFile(dto.getFileList(),productionTaskDetailRecord.getId());

        //推送消息给下一工序的用户有单可以抢了
        Long nextProductionProcessesId = productionTaskDetail.getNextProductionProcessesId();
        List<Long> userIds = productionTaskProcessesDetailService.listObject(ProductionTaskProcessesDetail::getUserId, q -> q
                .eq(ProductionTaskProcessesDetail::getProductionProcessesId, nextProductionProcessesId)
                .eq(ProductionTaskProcessesDetail::getProductionTaskId, productionTaskDetail.getProductionTaskId())
        );
        WebSocketPush.byUsers(PushTypeEnum.MESSAGE, userIds, "您有新的任务可接收！", PushBusinessTypeEnum.NEW_TASK.getType());
    }

    /**
     * 完工入库
     */
    @Override
    @DSTransactional
    public void productStorage(ProductionTaskDetailDto productionTaskDetailDto) {
        ProductionTaskDetail productionTaskDetail = getById(productionTaskDetailDto.getId());
        if (ObjectUtil.isEmpty(productionTaskDetail)) {
            throw new ServiceException("未查询到任务明细信息");
        }
        productionTaskDetail.setFinishStatus(1);//标记为完成
        productionTaskDetail.setFinishTime(new Date());//记录完成时间

        //设置当前工序状态完成
        productionTaskDetail.setProcessesStatus(3);
        this.updateById(productionTaskDetail);

        //创建任务明细记录
        ProductionTaskDetailRecord productionTaskDetailRecord = new ProductionTaskDetailRecord();
        productionTaskDetailRecord.setProductionTaskDetailId(productionTaskDetail.getId());
        productionTaskDetailRecord.setProductionProcessesId(productionTaskDetail.getProductionProcessesId());
        productionTaskDetailRecord.setPersonLiableId(SecurityUtils.getUserId());
        productionTaskDetailRecord.setReceivedTime(productionTaskDetail.getReceivedTime());
        productionTaskDetailRecordService.save(productionTaskDetailRecord);

        //保存附件
        ObsFileUtil.saveFile(productionTaskDetailDto.getFileList(),productionTaskDetailRecord.getId());

        StockWait stockWait = new StockWait();
        stockWait.setType(1);//入库
        stockWait.setStatus(0);//待入库
        stockWait.setBusinessType(StockWaitType.COMPLETION_IN.getDetailType());//
        stockWait.setBusinessId(productionTaskDetail.getId());//生产任务明细id
        stockWait.setBusinessCode(productionTaskDetail.getProductSn());//产品sn
        stockWaitService.save(stockWait);
        StockWaitDetails stockWaitDetails = new StockWaitDetails();
        stockWaitDetails.setBusinessDetailsId(productionTaskDetail.getId());
        stockWaitDetails.setStockWaitId(stockWait.getId());
        stockWaitDetails.setProductId(productionTaskDetail.getProductId());
        stockWaitDetails.setQuantity(BigDecimal.ONE);//生产任务明细入库 数量默认只有1
        stockWaitDetailsService.save(stockWaitDetails);
    }

    /**
     * 任务接收分页
     */
    @Override
    public Page<ProductionTaskDetailVo> receivePage(ProductionTaskDetailSelectDto dto) {
        IWrapper<ProductionTaskDetail> wrapper = getWrapper();

        //过滤工序用户列表里有自己的数据
        wrapper.eq("ptpd.user_id", SecurityUtils.getUserId());
        //过滤正在流转中的数据
        wrapper.eq("ptd", ProductionTaskDetail::getCirculationStatus, 1);
        //过滤未完工的数据
        wrapper.eq("ptd", ProductionTaskDetail::getFinishStatus, 0);

        return pageCom(wrapper, dto);
    }

    /**
     * 任务接收
     */
    @Override
    @DSTransactional
    public void receive(ProductionTaskDetailDto dto) {
        ProductionTaskDetail productionTaskDetail = getById(dto.getId());

        if (Objects.equals(0, productionTaskDetail.getCirculationStatus())) {
            throw new ServiceException("该任务已被他人接收，接收失败");
        }

        //修改流转状态
        productionTaskDetail.setCirculationStatus(0);
        //将负责人赋值为自己
        productionTaskDetail.setPersonLiableId(SecurityUtils.getUserId());
        //赋值接收时间
        productionTaskDetail.setReceivedTime(new Date());

        //设置当前工序状态进行中
        productionTaskDetail.setProcessesStatus(1);

        this.updateById(productionTaskDetail);
    }

    /**
     * 任务接收拒绝
     */
    @Override
    @DSTransactional
    public void rejection(ProductionTaskDetailDto productionTaskDetailDto) {
        ProductionTaskDetail productionTaskDetail = this.getById(productionTaskDetailDto.getId());
        //取消流转状态
        productionTaskDetail.setCirculationStatus(0);

        //回滚工序信息
        productionTaskDetail.setNextProductionProcessesId(productionTaskDetail.getProductionProcessesId());
        productionTaskDetail.setProductionProcessesId(productionTaskDetail.getPreviousProcessesId());

        //清空上一工序
        productionTaskDetail.setPreviousProcessesId(-1l);
        //查找下一工序相关信息 如果有则赋值
        ProductionTask productionTask = productionTaskService.getById(productionTaskDetail.getProductionTaskId());
        ProductionPlan productionPlan = productionPlanService.getById(productionTask.getProductionPlanId());
        WorkOrder workOrder = workOrderService.getById(productionPlan.getWorkOrderId());
        List<WorkOrderProductionProcesses> productionProcessesList = workOrderService.getProductionProcessesList(workOrder);
        //赋值上一工序id
        for (int i = 0; i < productionProcessesList.size(); i++) {
            WorkOrderProductionProcesses wopp = productionProcessesList.get(i);
            if (wopp.getId().equals(productionTaskDetail.getProductionProcessesId())) {
                //默认赋值空 如果有下一节点赋值下一节点id
                if ((i - 1) >=0) {
                    WorkOrderProductionProcesses nextProductionProcesses = productionProcessesList.get(i - 1);
                    productionTaskDetail.setPreviousProcessesId(nextProductionProcesses.getId());
                    break;
                }
            }
        }

        //设置当前工序状态驳回
        productionTaskDetail.setProcessesStatus(2);

        //回滚负责人信息
        productionTaskDetail.setPersonLiableId(productionTaskDetail.getCirculationUserId());

        this.updateById(productionTaskDetail);
    }


    @Override
    public ProductionTaskDetailVo snInfo(String sn) {
        Assert.notEmpty(sn, "sn不能为空");
        ProductionTaskDetailVo productionTaskDetailVo = baseMapper.snInfo(sn);
        Assert.notEmpty(productionTaskDetailVo, "查询不到sn信息");
        ProductInfo productInfo = productInfoService.getById(productionTaskDetailVo.getProductId());
        Assert.notEmpty(productInfo, "查询不到产品信息");
        productionTaskDetailVo.setProductName(productInfo.getName());
        return productionTaskDetailVo;
    }

    /**
     * 获取当前用户某一工序还在进行中的任务数
     */
    @Override
    public Long haveTaskCount(ProductionTaskDetailDto dto){
        //获取负责人为当前用户 工序为某一工序 未在流转（进行中） 未完成 的任务计数
        return this.count(q->q
                .eq(ProductionTaskDetail::getPersonLiableId,SecurityUtils.getUserId())
                .eq(ObjectUtil.isNotEmpty(dto.getProductionProcessesId()),ProductionTaskDetail::getProductionProcessesId,dto.getProductionProcessesId())
                .eq(ProductionTaskDetail::getCirculationStatus,0)
                .eq(ProductionTaskDetail::getFinishStatus,0)
        );
    }

    /**
     * 退回任务（任务回到任务列表可供别人接取）
     */
    @Override
    @DSTransactional
    public void revokeTask(ProductionTaskDetailDto dto){
        ProductionTaskDetail productionTaskDetail = getById(dto.getId());
        Assert.notEmpty(productionTaskDetail,"查询不到该任务信息");

        //修改流转状态
        productionTaskDetail.setCirculationStatus(1);
        //将负责人清空
        productionTaskDetail.setPersonLiableId(-1L);

        //设置当前工序状态驳回
        productionTaskDetail.setProcessesStatus(0);

        this.updateById(productionTaskDetail);
    }

    /**
     * 生产进度
     */
    @Override
    public Page<ProductionTaskDetailVo> taskProgressPage(ProductionTaskDetailSelectDto dto){
        IWrapper<ProductionTaskDetail> wrapper = IWrapper.getWrapper();
        wrapper.keyword(dto.getKeyword(),
                new SqlField("t1.contractCode"),
                new SqlField("t1.workOrderCode"),
                new SqlField("t1.productionPlanCode"),
                new SqlField("t1.productionTaskCode"),
                new SqlField("t1.productCode"),
                new SqlField("t1.productName"),
                new SqlField("t1",ProductionTaskDetail::getProductSn),
                new SqlField("t1.due_date"),
                new SqlField("t1.productionProcessesName"),
                new SqlField("t1.personLiableName"),
                new SqlField("t1",ProductionTaskDetail::getFinishTime)
        );
        //高级检索
        wrapper.like("t1.contractCode",dto.getContractCode());
        wrapper.like("t1.workOrderCode",dto.getWorkOrderCode());
        wrapper.like("t1.productionPlanCode",dto.getProductionPlanCode());
        wrapper.like("t1.productionTaskCode",dto.getProductionTaskCode());
        wrapper.like("t1.productCode",dto.getProductCode());
        wrapper.like("t1.productName",dto.getProductName());
        wrapper.like("t1",ProductionTaskDetail::getProductSn,dto.getProductSn());
        wrapper.eq("t1.due_date",dto.getDueDate());
        wrapper.like("t1.productionProcessesName",dto.getProductionProcessesName());
        wrapper.like("t1.personLiableName",dto.getPersonLiableName());
        wrapper.eq("t1",ProductionTaskDetail::getFinishStatus,dto.getGetFinishStatus());
//        wrapper.eq("t1",ProductionTaskDetail::getFinishTime,dto.getGetFinishTime());

        //完成时间包含时分秒需要将时间特殊处理
        if(ObjectUtil.isNotEmpty(dto.getGetFinishTime())) {
            Date beginTime = DateUtil.beginOfDay(DateUtil.parse(dto.getGetFinishTime()));
            Date endTime = DateUtil.endOfDay(DateUtil.parse(dto.getGetFinishTime()));
            wrapper.between("t1", ProductionTaskDetail::getFinishTime, beginTime, endTime);
        }


        wrapper.orderByDesc("t1",ProductionTaskDetail::getId);

        Page<ProductionTaskDetailVo> page = baseMapper.taskProgressPage(dto.getPage(),wrapper);
        return page;
    }

    /**
     * 生产任务明细
     */
    @Override
    public ProductionTaskDetail detail(Long id) {
        ProductionTaskDetail productionTaskDetail = this.getById(id);
        ProductionTask productionTask = productionTaskService.getById(productionTaskDetail.getProductionTaskId());
        ProductionPlan productionPlan = productionPlanService.getById(productionTask.getProductionPlanId());
        WorkOrder workOrder = workOrderService.getById(productionPlan.getWorkOrderId());

        List<Long> processesIds = Arrays.asList(productionTaskDetail.getPreviousProcessesId(),
                productionTaskDetail.getProductionProcessesId(),
                productionTaskDetail.getNextProductionProcessesId()
        );

        Map<Long,String> productionProcessesMap;
        if (Objects.equals(workOrder.getIsCustomized(),1)){
            //定制
            productionProcessesMap = workOrderProductionProcessesService.mapKV(WorkOrderProductionProcesses::getId,
                    WorkOrderProductionProcesses::getName, q -> q.in(WorkOrderProductionProcesses::getId, processesIds));
        }else {
            //非定制
            productionProcessesMap = productionProcessesService.mapKV(ProductionProcesses::getId, ProductionProcesses::getName,
                    q -> q.in(ProductionProcesses::getId, processesIds));
        }

        ProductionTaskDetailVo productionTaskDetailVo = BeanUtil.copyProperties(productionTaskDetail, ProductionTaskDetailVo.class);

        productionTaskDetailVo.setProductionProcessesName(productionProcessesMap.get(productionTaskDetail.getProductionProcessesId()));
        productionTaskDetailVo.setNextProductionProcessesName(productionProcessesMap.get(productionTaskDetail.getNextProductionProcessesId()));
        productionTaskDetailVo.setPreviousProcessesName(productionProcessesMap.get(productionTaskDetail.getPreviousProcessesId()));

        List<ProductionTaskDetailRecord> productionTaskDetailRecordList = productionTaskDetailRecordService.list(q -> q
                .eq(ProductionTaskDetailRecord::getProductionTaskDetailId, id)
                .orderByDesc(ProductionTaskDetailRecord::getCreateTime)
        );
        List<ProductionTaskDetailRecordVo> productionTaskDetailRecordVos = BeanUtil.copyToList(productionTaskDetailRecordList, ProductionTaskDetailRecordVo.class);
        productionTaskDetailVo.setProductionTaskDetailRecordList(productionTaskDetailRecordVos);

        return productionTaskDetailVo;
    }

}
